from .socp import Model as SOCModel
from .lp import LinConstr, Bounds, CvxConstr, ConeConstr
from .lp import Vars, VarSub, Affine, Convex
from .lp import DecRule, DecRuleSub
from .lp import RoAffine, RoConstr
from .lp import Solution
from .subroutines import *
import numpy as np
from scipy.sparse import csr_matrix
from collections import Iterable
from .lpg_solver import solve as def_sol


class Model:
    """
    The Model class creates an object of robust optimization models
    """

    def __init__(self, name=None):

        self.rc_model = SOCModel(mtype='R')
        self.sup_model = SOCModel(nobj=True, mtype='S')

        self.all_constr = []

        self.obj = None
        self.obj_support = None
        self.sign = 1

        self.primal = None
        self.dual = None
        self.solution = None
        self.pupdate = True
        self.dupdate = True

        self.solution = None

        self.name = name

    def reset(self):

        self.all_constr = []
        self.pupdate = True
        self.dupdate = True
        self.primal = None
        self.dual = None
        self.rc_model.reset()

    def dvar(self, shape=(1,), vtype='C', name=None, aux=False):
        """
        Returns an array of decision variables with the given shape
        and variable type.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        vtype : {'C', 'B', 'I'}
            Type of the decision variables. 'C' means continuous; 'B'
            means binary, and 'I" means integer.
        name : str
            Name of the variable array
        aux : leave it unspecified.

        Returns
        -------
        new_var : rsome.lp.Vars
            An array of new decision variables
        """

        new_var = self.rc_model.dvar(shape, vtype, name, aux)
        return new_var

    def rvar(self, shape=(1,), name=None):

        """
        Returns an array of random variables with the given shape.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        name : str
            Name of the variable array

        Returns
        -------
        new_var : rsome.lp.Vars
            An array of new random variables
        """

        new_var = self.sup_model.dvar(shape, 'C', name)
        return new_var

    def ldr(self, shape=(1,), name=None):

        """
        Returns an array with the given shape of linear decision rule
        variables.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        name : str
            Name of the variable array

        Returns
        -------
        new_var : rsome.ro.DecRule
            An array of new linear decision rule variables
        """

        new_ldr = DecRule(self, shape, name)
        return new_ldr

    def min(self, obj):
        """
        Minimize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if obj.size > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def max(self, obj):
        """
        Maximize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if obj.size > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def minmax(self, obj, *args):

        """
        Minimize the maximum objective value over the given uncertainty set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        *args
            Constraints or collections of constraints of random variables
            used for defining the uncertainty set

        Notes
        -----
        The uncertainty set defined for the objective function is considered
        the default uncertainty set for the robust model.
        """

        if np.prod(obj.shape) > 1:
            raise ValueError('Incorrect function dimension.')

        constraints = []
        for items in args:
            if isinstance(items, Iterable):
                constraints.extend(list(items))
            else:
                constraints.append(items)

        sup_model = self.sup_model
        sup_model.reset()
        for item in constraints:
            if item.model is not sup_model:
                raise SyntaxError('Models mismatch.')
            sup_model.st(item)

        self.obj = obj
        self.obj_support = sup_model.do_math(primal=False)
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def maxmin(self, obj, *args):

        """
        Maximize the minimum objective value over the given uncertainty set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        *args
            Constraints or collections of constraints of random variables
            used for defining the uncertainty set

        Notes
        -----
        The uncertainty set defined for the objective function is considered
        the default uncertainty set for the robust model.
        """

        if np.prod(obj.shape) > 1:
            raise ValueError('Incorrect function dimension.')

        constraints = []
        for items in args:
            if isinstance(items, Iterable):
                constraints.extend(list(items))
            else:
                constraints.append(items)

        sup_model = self.sup_model
        sup_model.reset()
        for item in constraints:
            if item.model is not sup_model:
                raise SyntaxError('Models mismatch.')
            sup_model.st(item)

        self.obj = obj
        self.obj_support = sup_model.do_math(primal=False)
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def st(self, *arg):

        for constr in arg:
            if isinstance(constr, Iterable):
                for item in constr:
                    self.st(item)

            elif isinstance(constr, (LinConstr, Bounds, CvxConstr, ConeConstr)):
                if (constr.model is not self.rc_model) or \
                        (constr.model.mtype != 'R'):
                    raise ValueError('Models mismatch.')
                self.all_constr.append(constr)
            elif isinstance(constr, RoConstr):
                if (constr.dec_model is not self.rc_model) or \
                        (constr.rand_model is not self.sup_model):
                    raise ValueError('Models mismatch.')
                sense = (constr.sense[0] if isinstance(constr.sense,
                                                       np.ndarray)
                         else constr.sense)
                if sense == 0:
                    self.all_constr.append(constr)
                else:
                    left = RoAffine(constr.raffine, constr.affine,
                                    constr.rand_model)
                    right = RoAffine(-constr.raffine, -constr.affine,
                                     constr.rand_model)
                    self.all_constr.append(RoConstr(left, sense=0))
                    self.all_constr.append(RoConstr(right, sense=0))
            else:
                raise TypeError('Unknown type of constraints')

    def do_math(self, primal=True):

        if primal:
            if self.primal is not None and not self.pupdate:
                return self.primal
        else:
            if self.dual is not None and not self.dupdate:
                return self.dual
            else:
                self.do_math(primal=True)
                return self.rc_model.do_math(False, obj=True)

        self.rc_model.reset()
        if isinstance(self.obj, (Vars, VarSub, Affine, Convex)):
            self.rc_model.obj = self.obj
            self.rc_model.sign = self.sign
            more_roc = []
        elif isinstance(self.obj, RoAffine):
            obj_constr = (self.rc_model.vars[0] >= self.sign * self.obj)
            obj_constr.support = self.obj_support
            more_roc = [obj_constr]
        else:
            raise TypeError('Incorrect type for the objective function.')

        for constr in self.all_constr + more_roc:
            if isinstance(constr, (LinConstr, Bounds, CvxConstr)):
                self.rc_model.st(constr)
            if isinstance(constr, RoConstr):
                if constr.support:
                    rc_constrs = constr.le_to_rc()
                else:
                    rc_constrs = constr.le_to_rc(self.obj_support)
                for rc_constr in rc_constrs:
                    self.rc_model.st(rc_constr)

        formula = self.rc_model.do_math(primal, obj=True)

        if primal:
            self.primal = formula
            self.pupdate = False
        else:
            self.dual = formula
            self.dupdate = False

        return formula

    def solve(self, solver=None, display=True, export=False, params={}):
        """
        Solve the model with the selected solver interface.

        Parameters
        ----------
            solver : {None, lpg_solver, grb_solver, msk_solver}
                Solver interface used for model solution. Use default solver
                if solver=None.
            display : bool
                Display option of the solver interface.
            export : bool
                Export option of the solver interface. A standard model file
                is generated if the option is True.
            params : dict
                A dictionary that specifies parameters of the selected solver.
                So far the argument only applies to Gurobi and MOSEK.
        """

        if solver is None:
            solution = def_sol(self.do_math(), display, export, params)
        else:
            solution = solver.solve(self.do_math(), display, export, params)

        if isinstance(solution, Solution):
            self.rc_model.solution = solution
        else:
            if solution is None:
                self.rc_model.solution = None
            else:
                x = solution.x
                self.rc_model.solution = Solution(x[0], x, solution.status)

        self.solution = self.rc_model.solution

    def get(self):

        if self.rc_model.solution is None:
            raise SyntaxError('The model is unsolved or no feasible solution.')
        return self.sign * self.rc_model.solution.objval



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='rsome',
     version='0.1.4',
     author="Peng Xiong, Zhi Chen, and Melvyn Sim",
     author_email="xiongpengnus@gmail.com",
     description="Robust Stochastic Optimization Made Easy",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/XiongPengNUS/rsome",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "Operating System :: OS Independent",
     ],
 )



"""
This module is used as an interface to call the Gurobi solver for solving
(mixed-integer) linear or second-order cone programs.
"""

import gurobipy as gp
import numpy as np
import warnings
import time
from .socp import SOCProg
from .lp import Solution


def solve(formula, display=True, export=False, params={}):

    nv = formula.linear.shape[1]
    vtype = list(formula.vtype)

    grb = gp.Model()
    x = grb.addMVar(nv, lb=formula.lb, ub=formula.ub, vtype=vtype)

    indices_eq = (formula.sense == 1)
    indices_ineq = (formula.sense == 0)
    linear_eq = formula.linear[indices_eq, :]
    linear_ineq = formula.linear[indices_ineq, :]
    const_eq = formula.const[indices_eq]
    const_ineq = formula.const[indices_ineq]
    if len(indices_eq) > 0:
        grb.addMConstrs(linear_eq, x, '=', const_eq)
    if len(indices_ineq) > 0:
        grb.addMConstrs(linear_ineq, x, '<', const_ineq)

    if isinstance(formula, SOCProg):
        for constr in formula.qmat:
            index_right = constr[0]
            index_left = constr[1:]
            A = np.eye(len(index_left))
            grb.addConstr(x[index_left] @ A @ x[index_left] <=
                          x[index_right] @ x[index_right])

    grb.setObjective(formula.obj @ x)

    grb.setParam('LogToConsole', 0)
    try:
        for param, value in params.items():
            grb.setParam(param, value)
    except (TypeError, ValueError):
        raise ValueError('Incorrect parameters or values.')
    if display:
        print('Being solved by Gurobi...', flush=True)
        time.sleep(0.2)
    grb.optimize()
    if display:
        print('Solution status: {0}'.format(grb.Status))
        print('Running time: {0:0.4f}s'.format(grb.Runtime))

    if export:
        grb.write("out.lp")

    try:
        solution = Solution(grb.ObjVal, grb.getAttr('X'), grb.Status)
    except AttributeError:
        warnings.warn('No feasible solution can be found.')
        solution = None

    return solution



"""
This module is used as an interface to call the CPLEX solver for solving 
(mixed-integer) linear or second-order cone programs.
"""

import cplex
import numpy as np
import warnings
import time
from .lp import Solution


def solve(formula, display=True, export=False, params={}):

    cpx = cplex.Cplex()

    obj = formula.obj.flatten()
    linear = formula.linear
    row = linear.shape[0]
    spmat = [[linear.indices[linear.indptr[i]:linear.indptr[i + 1]].tolist(),
              linear.data[linear.indptr[i]:linear.indptr[i + 1]].tolist()]
              for i in range(row)]
    sense = ['E' if s == 1 else 'L' for s in formula.sense]
    const = formula.const
    ub = formula.ub
    lb = formula.lb
    vtype = [cpx.variables.type.integer if vt == 'I' else
             cpx.variables.type.binary if vt == 'B' else
             cpx.variables.type.continuous for vt in formula.vtype]

    if all(np.array(vtype) == cpx.variables.type.continuous):
        cpx.variables.add(obj=formula.obj.flatten(),
                          lb=formula.lb, ub=formula.ub)
    else:
        cpx.variables.add(obj=formula.obj.flatten(),
                          lb=formula.lb, ub=formula.ub, types=vtype)
    cpx.linear_constraints.add(lin_expr=spmat,
                               senses=sense, rhs=formula.const)

    for cone in formula.qmat:
        cone_data = [-1] + [1] * (len(cone) - 1)
        cone = [int(index) for index in cone]
        q = cplex.SparseTriple(ind1=cone, ind2=cone, val=cone_data)
        cpx.quadratic_constraints.add(quad_expr=q)

    if display:
        print('Being solved by CPLEX...', flush=True)
        time.sleep(0.2)

    cpx.set_results_stream(None)
    cpx.set_warning_stream(None)
    try:
        for param, value in params.items():
            text = 'cpx.parameters.' + param + '.set({0})'.format(value)
            eval(text)
    except (TypeError, ValueError):
        raise ValueError('Incorrect parameters or values.')

    t0 = time.time()
    cpx.solve()
    stime = time.time() - t0
    status = cpx.solution.get_status()
    if display:
        print('Solution status: {0}'.format(status))
        print('Running time: {0:0.4f}s'.format(stime))

    if status in [1, 6, 10, 11, 12, 13,
                  21, 22,
                  101, 102, 105, 107, 109, 111, 113, 116]:
        obj_val = cpx.solution.get_objective_value()
        x_sol = np.array(cpx.solution.get_values())
        solution = Solution(obj_val, x_sol, status)
    else:
        warnings.warn('No feasible solution can be found.')
        solution = None

    return solution



"""
This module is used as an interface to call the scipy.optimize.linprog
solver (default solver) for solving linear programming problems.
"""

import scipy.optimize as opt
import numpy as np
import warnings
import time

def solve(formula, display=True, export=False, params={}):

    if export:
        warnings.warn('Cannot export model by the linprog() function. ')
    try:
        if formula.qmat:
            warnings.warn('SOC constriants are ignored in the LP solver. ')
    except AttributeError:
        pass
    if any(np.array(formula.vtype) != 'C'):
        warnings.warn('Integrality constraints are ignored in the LP solver. ')

    nv = formula.linear.shape[1]
    vtype = list(formula.vtype)

    indices_eq = (formula.sense == 1)
    indices_ineq = (formula.sense == 0)
    linear_eq = formula.linear[indices_eq, :]
    linear_ineq = formula.linear[indices_ineq, :]
    const_eq = formula.const[indices_eq]
    const_ineq = formula.const[indices_ineq]

    if len(indices_ineq) == 0:
        linear_ineq = None
        const_ineq = None
    if len(indices_eq) == 0:
        linear_eq = None
        const_eq = None

    bounds = [(lb, ub) for lb, ub in zip(formula.lb, formula.ub)]

    default = {'maxiter': 1000000000,
               'sparse': True}

    if display:
        print('Being solved by the default LP solver...', flush=True)
        time.sleep(0.2)
    t0 = time.time()
    res = opt.linprog(formula.obj, A_ub=linear_ineq, b_ub=const_ineq,
                      A_eq=linear_eq, b_eq=const_eq,
                      bounds=bounds, options=default)
    stime = time.time() - t0
    if display:
        print('Solution status: {0}'.format(res.status))
        print('Running time: {0:0.4f}s'.format(stime))

    if res.status == 1:
        warnings.warn('Iteration limit reached. ')
    if res.status == 2:
        warnings.warn('Problem appears to be infeasible. ')
    if res.status == 3:
        warnings.warn('Problem appears to be unbounded. ')
    if res.status == 4:
        warnings.warn('Numerical difficulties encountered. ')

    return res



import numpy as np
import scipy.sparse as sp
from numbers import Real
from scipy.sparse import csr_matrix
from collections import Iterable


def flat(a_list):
    flat_list = []
    for item in a_list:
        if isinstance(item, Iterable):
            flat_list.extend(flat(item))
        else:
            flat_list.append(item)

    return flat_list


def sparse_mul(ndarray, affine):

    array_ones = np.ones(ndarray.shape, dtype='int')
    affine_ind = np.arange(affine.size, dtype='int').reshape(affine.shape)
    affine_ones = np.ones(affine.shape, dtype='int')

    index = (array_ones * affine_ind).flatten()
    values = (ndarray * affine_ones).flatten()

    size = index.size

    return csr_matrix((values, (np.arange(size), index)),
                      shape=(size, affine.size))


def sp_matmul(ndarray, affine, shape):

    size = np.prod(shape)

    left = (ndarray.reshape((ndarray.size//ndarray.shape[-1],
                             ndarray.shape[-1])) if ndarray.ndim > 1
            else ndarray.reshape((1, ndarray.size)))
    right = (affine if len(affine.shape) > 1
             else affine.reshape((affine.size, 1)))

    row, inner = left.shape
    column = right.shape[-1]

    index = np.arange(affine.size).reshape(affine.shape)
    indim = index.ndim
    index = np.transpose(np.tile(index, (1, row)),
                         axes=list(range(indim-2)) + [indim-1,
                                                      indim-2]).flatten()
    index = np.tile(index, (size*inner//index.size, ))

    data = np.tile(ndarray, (1, column)).flatten()
    data = np.tile(data, size*inner//data.size)

    indptr = [inner*i for i in range(size+1)]

    return csr_matrix((data, index, indptr), shape=[size, affine.size])


def sp_lmatmul(ndarray, affine, shape):

    size = int(np.prod(shape))

    left = (affine.reshape((affine.size // affine.shape[-1],
                            affine.shape[-1])) if len(affine.shape) > 1
            else affine.reshape((1, affine.size)))
    right = (ndarray if ndarray.ndim > 1
             else ndarray.reshape((ndarray.size, 1)))

    row, inner = left.shape
    column = right.shape[-1]

    index = np.tile(np.arange(left.size).reshape(left.shape),
                    (1, column)).flatten()
    index = np.tile(index, (size * inner // index.size))
    andim = right.ndim
    data = np.tile(np.transpose(right,
                                axes=list(range(andim - 2)) + [andim - 1,
                                                               andim - 2]),
                   (size * inner // ndarray.size, 1)).flatten()
    indptr = [inner * i for i in range(size + 1)]

    return csr_matrix((data, index, indptr), shape=[size, affine.size])


def sp_trans(affine):

    index = np.arange(affine.size).reshape(affine.shape).T.flatten()
    indptr = np.arange(affine.size + 1)
    data = np.ones(affine.size)

    return csr_matrix((data, index, indptr), shape=[affine.size, affine.size])


def index_array(shape):

    if isinstance(shape, tuple):
        size = np.prod(shape)
    else:
        size = int(shape)
        shape = int(shape),

    return np.arange(size, dtype=int).reshape(shape)


def sparse_array(shape):

    shape = shape if isinstance(shape, tuple) else (int(shape), )
    size = np.prod(shape).item()
    sparse = csr_matrix(([1.0]*size, np.arange(size, dtype=int),
                         np.arange(size+1, dtype=int)), (size, size))
    elements = [sparse[i, :] for i in range(size)]

    return np.array(elements).reshape(shape)


def array_to_sparse(array):

    size = array.size
    all_items = list(array.reshape((size, )))

    indices = np.concatenate(tuple(item.indices for item in all_items))
    data = np.concatenate(tuple([item.data for item in all_items]))
    indices = indices[data != 0]
    data = data[data != 0]

    indptr = np.zeros(size+1)
    for i in range(size):
        indptr[1+i] = indptr[i] + all_items[i].count_nonzero()

    return csr_matrix((data, indices, indptr), (size, all_items[0].shape[1]))


def sv_to_csr(array):

    size = array.size
    all_items = list(array.reshape((size, )))

    indices = np.concatenate(tuple(item.index for item in all_items))
    data = np.concatenate(tuple([item.value for item in all_items]))
    nonzero = (data != 0)
    indices = indices[nonzero]
    data = data[nonzero]

    indptr = np.zeros(size+1)
    for i in range(size):
        zero_counts = all_items[i].value.count(0)
        indptr[1+i] = indptr[i] + len(all_items[i].value) - zero_counts

    return csr_matrix((data, indices, indptr), (size, all_items[0].nvar))


def check_numeric(array):

    array = np.array(array.todense()) if sp.issparse(array) else array
    array = np.array([array]) if not isinstance(array, np.ndarray) else array

    if isinstance(array, np.ndarray):
        if not isinstance(array.flat[0], np.number):
            raise TypeError('Incorrect data type of arrays.')
    elif not isinstance(array, Real):
        raise TypeError('Incorrect data type of arrays.')

    return array


def add_linear(left, right):

    if left.shape[1] > right.shape[1]:
        right.resize((right.shape[0], left.shape[1]))
    elif right.shape[1] > left.shape[1]:
        left.resize((left.shape[0], right.shape[1]))

    return left + right


def event_dict(event_set):

    output = {}
    count = 0
    for item in event_set:
        for element in item:
            output[element] = count

        count += 1

    return output


def comb_set(s1, s2):

    d1 = event_dict(s1)
    d2 = event_dict(s2)

    dc = {item: str(d1[item]) + '-' + str(d2[item])
          for item in range(len(d1))}

    values = []
    output = []
    for key in dc:
        if dc[key] in values:
            index = values.index(dc[key])
            output[index].append(key)
        else:
            output.append([key])
            values.append(dc[key])

    return output


def norm(affine, degree=2):
    """
    Return the first, second, or infinity norm of an array.

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array. The array must be a vector, i.e., the size of the
        array must be the same as the largest dimension.
    degree : {1, 2, numpy.inf}, optional
        Order of the norm function. It can only be 1, 2, or infinity.

    Returns
    -------
    n : Convex
        The norm of the given array
    """

    return affine.norm(degree)


def square(affine):
    """
    Return the element-wise square of an array.

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    Returns
    -------
    n : Convex
        The element-wise squares of the given array
    """

    return affine.to_affine().square()


def sumsqr(affine):
    """
    Return the sum of squares of an array.

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array. The array must be a vector, i.e., the size of the
        array must be the same as the largest dimension.

    Returns
    -------
    n : Convex
        The sum of squares of the given array
    """

    return affine.to_affine().sumsqr()


def E(expr):
    """
    The notion of the expectation of random variables and the worst-case
    expected values

    Notes
    -----
    This function is used to denote
    1. the expected value of an random variable when specifying the
    uncertainty set of expectations.
    2. the worst-case expected value of an affine function involving random
    variables.
    """

    return expr.E



from .socp import Model as SOCModel
from .ro import Model as ROModel
from .lp import LinConstr, ConeConstr, CvxConstr
from .lp import Vars, Affine
from .lp import RoAffine, RoConstr
from .lp import DecVar, RandVar, DecLinConstr, DecCvxConstr
from .lp import DecRoConstr
from .lp import Scen
from .lp import Solution
from .lpg_solver import solve as def_sol
from .subroutines import *
import numpy as np
import pandas as pd
import warnings
from collections import Sized, Iterable


class Model:
    """
    Returns a model object with the given number of scenarios.

    Parameters
    ----------
    scens : int or array-like objects
        The number of scenarios, if it is an integer. It could also be
        an array of scenario indices.
    name : str
        Name of the model

    Returns
    -------
    model : rsome.dro.Model
        A model object
    """

    def __init__(self, scens=1, name=None):

        self.ro_model = ROModel()
        self.vt_model = SOCModel(mtype='V')
        self.sup_model = self.ro_model.sup_model
        self.exp_model = SOCModel(nobj=True, mtype='E')
        self.pro_model = SOCModel(nobj=True, mtype='P')

        self.obj_ambiguity = None

        if isinstance(scens, int):
            num_scen = scens
            series = pd.Series(np.arange(num_scen).astype(int))
        elif isinstance(scens, Sized):
            num_scen = len(scens)
            series = pd.Series(np.arange(num_scen).astype(int), index=scens)
        else:
            raise ValueError('Incorrect scenarios.')
        self.num_scen = num_scen
        self.series_scen = series

        self.dec_vars = [DecVar(self, self.vt_model.vars[0])]
        self.rand_vars = []
        self.all_constr = []
        self.var_ev_list = None
        # self.affadapt_mat = None

        pr = self.pro_model.dvar(num_scen, name='probabilities')
        self.p = pr

        self.obj = None
        self.sign = 1

        self.primal = None
        self.dual = None
        self.solution = None
        self.pupdate = True
        self.dupdate = True

        self.solution = None

        self.name = name

    def rvar(self, shape=(1,), name=None):

        sup_var = self.sup_model.dvar(shape, 'C', name)
        exp_var = self.exp_model.dvar(shape, 'C', name)
        rand_var = RandVar(sup_var, exp_var)
        self.rand_vars.append(rand_var)

        return rand_var

    def dvar(self, shape=(1,), vtype='C', name=None):
        """
        Returns an array of decision variables with the given shape
        and variable type.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        vtype : {'C', 'B', 'I'}
            Type of the decision variables. 'C' means continuous; 'B'
            means binary, and 'I" means integer.
        name : str
            Name of the variable array

        Returns
        -------
        new_var : rsome.lp.DecVar
            An array of new decision variables
        """

        dec_var = self.vt_model.dvar(shape, vtype, name)
        dec_var = DecVar(self, dec_var, name=name)
        self.dec_vars.append(dec_var)

        return dec_var

    def ambiguity(self):
        """
        Returns an event-wise ambiguity set with the given number of scenarios
        decalred in creating the DRO model.
        """

        if self.all_constr:
            raise SyntaxError('Ambiguity set must be specified ' +
                              'before defining constraints.')

        return Ambiguity(self)

    def bou(self, *args):

        if self.num_scen != 1:
            raise ValueError('The uncertainty set can only be applied '
                             'to a one-scenario model')

        bou_set = self.ambiguity()
        for arg in args:
            if arg.model is not self.sup_model:
                raise ValueError('Constraints are not for this support.')

        bou_set.sup_constr = [tuple(args)]

        return bou_set

    def wks(self, *args):

        if self.num_scen != 1:
            raise ValueError('The WKS ambiguity set can only be applied '
                             'to a one-scenario model')

        wks_set = self.ambiguity()
        sup_constr = []
        exp_constr = []
        for arg in args:
            if arg.model is self.sup_model:
                sup_constr.append(arg)
            elif arg.model is self.exp_model:
                exp_constr.append(arg)
            else:
                raise ValueError('Constraints are not defined for the '
                                 'ambiguity support.')

        wks_set.sup_constr = [tuple(sup_constr)]
        wks_set.exp_constr = [tuple(exp_constr)]
        wks_set.exp_constr_indices = [np.array([0], dtype=np.int32)]

        return wks_set

    def adro(self, *args):

        return self.wks(*args)

    def rule_var(self):

        if self.var_ev_list is not None:
            return self.var_ev_list

        total = sum(dvar.size*len(dvar.event_adapt)
                    for dvar in self.dec_vars)
        vtype = ''.join([dvar.vtype * dvar.size * len(dvar.event_adapt)
                         if len(dvar.vtype) == 1
                         else dvar.vtype * len(dvar.event_adapt)
                         for dvar in self.dec_vars])
        var_const = self.ro_model.dvar(total, vtype=vtype)

        count = 0
        for dvar in self.dec_vars:
            dvar.ro_first = count
            count += dvar.size*len(dvar.event_adapt)

        num_scen = self.num_scen
        self.var_ev_list = []
        for s in range(num_scen):
            start = 0
            index = []
            total_size = 0
            total_col = 0
            for dvar in self.dec_vars:
                edict = event_dict(dvar.event_adapt)
                size = dvar.size
                index.extend(list(start + size * edict[s]
                                  + np.arange(size, dtype=int)))

                start += size * len(dvar.event_adapt)
                total_size += size
                total_col += size * len(dvar.event_adapt)

            self.var_ev_list.append(var_const[index].to_affine())

        if self.sup_model.vars:

            adapt_list = [dvar.rand_adapt if dvar.rand_adapt is not None else
                          np.zeros((dvar.size, self.sup_model.vars[-1].last))
                          for dvar in self.dec_vars]
            depend_mat = np.concatenate(adapt_list, axis=0)
            if depend_mat.sum() > 0:

                num_depend = np.array([int(dvar.rand_adapt.sum())
                                       if dvar.rand_adapt is not None else 0
                                       for dvar in self.dec_vars])
                scen_depend = np.array([len(dvar.event_adapt)
                                        for dvar in self.dec_vars])
                total_depend = num_depend @ scen_depend
                each_depend = num_depend.sum()
                var_linear = self.ro_model.dvar(total_depend)
                nz_rows = np.where(depend_mat.flatten())[0]
                for s in range(num_scen):
                    start = 0
                    index = []
                    total_var = 0
                    for dvar, num, scen in zip(self.dec_vars,
                                               num_depend, scen_depend):
                        if num == 0:
                            continue
                        edict = event_dict(dvar.event_adapt)
                        index.extend(list(start + num * edict[s]
                                     + np.arange(num, dtype=int)))
                        start += num * scen
                        total_var += num

                    depend_affine = var_linear[index].to_affine()

                    nz_cols = depend_affine.linear.indices
                    last = depend_affine.linear.shape[1]

                    ra_linear = sp.csr_matrix(([1.0] * each_depend,
                                               (nz_rows, nz_cols)),
                                              shape=(depend_mat.size, last))
                    ra_const = np.zeros(depend_mat.shape)
                    raffine = Affine(self.ro_model.rc_model,
                                     ra_linear, ra_const)
                    self.var_ev_list[s] = RoAffine(raffine,
                                                   self.var_ev_list[s],
                                                   self.ro_model.sup_model)

        return self.var_ev_list

    def min(self, obj):
        """
        Minimize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if obj.size > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def max(self, obj):
        """
        Maximize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if obj.size > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def minsup(self, obj, ambset):
        """
        Minimize the worst-case expected objective value over the given
        ambiguity set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        ambset : Ambiguity
            The ambiguity set defined for the worst-case expectation

        Notes
        -----
        The ambiguity set defined for the objective function is considered
        the default ambiguity set for the distritionally robust model.
        """

        if np.prod(obj.shape) > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.obj_ambiguity = ambset
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def maxinf(self, obj, ambset):

        """
        Maximize the worst-case expected objective value over the given
        ambiguity set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        ambset : Ambiguity
            The ambiguity set defined for the worst-case expectation

        Notes
        -----
        The ambiguity set defined for the objective function is considered
        the default ambiguity set for the distritionally robust model.
        """

        if np.prod(obj.shape) > 1:
            raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.obj_ambiguity = ambset
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def st(self, *arg):

        for constr in arg:
            if isinstance(constr, Iterable):
                for item in constr:
                    self.st(item)
            else:
                self.all_constr.append(constr)

    def do_math(self, primal=True):

        if primal:
            if self.primal is not None and not self.pupdate:
                return self.primal
        else:
            if self.dual is not None and not self.dupdate:
                return self.dual
            else:
                self.do_math(primal=True)
                return self.ro_model.do_math(False)

        self.ro_model.reset()
        self.rule_var()

        # Event-wise objective function
        self.ro_model.min(self.ro_model.rc_model.vars[1][0].to_affine())
        sign = self.sign
        constr = (self.dec_vars[0] >= self.obj * sign)
        if isinstance(constr, DecCvxConstr):
            ro_constr_list = self.ro_to_roc(constr)
        elif constr.ctype == 'R':
            ro_constr_list = self.ro_to_roc(constr)
        elif constr.ctype == 'E':
            ro_constr_list = self.dro_to_roc(constr)
        else:
            raise SyntaxError('Syntax error.')

        self.ro_model.st(ro_constr_list)

        # Event-wise Constraints
        for constr in self.all_constr:
            if isinstance(constr, DecCvxConstr):
                ro_constr_list = self.ro_to_roc(constr)
            elif constr.ctype == 'R':
                ro_constr_list = self.ro_to_roc(constr)
            elif constr.ctype == 'E':
                ro_constr_list = self.dro_to_roc(constr)
            else:
                raise SyntaxError('Syntax error')
            self.ro_model.st(ro_constr_list)

        formula = self.ro_model.do_math(primal)
        if primal:
            self.primal = formula
            self.pupdate = False
        else:
            self.dual = formula
            self.dupdate = False

        return formula

    def ro_to_roc(self, constr):

        drule_list = self.rule_var()
        num_var = self.vt_model.vars[-1].last

        ro_constr = []
        # for event in constr.event_adapt:
        for s in range(self.num_scen):
            drule = drule_list[s]
            if isinstance(constr, DecRoConstr):
                is_equal = (all(constr.sense) if
                            isinstance(constr.sense, Iterable) else
                            constr.sense == 1)
                if is_equal:
                    roaffine = RoAffine(constr.raffine, constr.affine,
                                        constr.rand_model)
                    left = DecRoConstr(roaffine, 0,
                                       constr.event_adapt, constr.ctype)
                    left.ambset = constr.ambset
                    right = DecRoConstr(-roaffine, 0,
                                        constr.event_adapt, constr.ctype)
                    right.ambset = constr.ambset

                    return self.ro_to_roc(left) + self.ro_to_roc(right)

                raf_linear, aff_linear = (constr.raffine.linear[:, :num_var],
                                          constr.affine.linear[:, :num_var])

                row_ind = np.unique(raf_linear.indices)
                if isinstance(drule, RoAffine):
                    if len(row_ind) > 0:
                        if (drule.raffine[row_ind].linear.nnz > 0 or
                                np.any(drule.raffine[row_ind].const)):
                            raise SyntaxError('Incorrect affine expressions.')

                    raffine = raf_linear @ drule.affine
                    raffine = (raffine.reshape(constr.raffine.shape)
                               + constr.raffine.const)
                    roaffine = RoAffine(raffine, np.zeros(aff_linear.shape[0]),
                                        self.sup_model)
                    this_const = constr.affine.const
                    this_const = this_const.flatten()
                    roaffine = roaffine + aff_linear@drule + this_const
                    ew_constr = RoConstr(roaffine, constr.sense)

                elif isinstance(drule, Affine):
                    raffine = raf_linear @ drule
                    raffine = (raffine.reshape(constr.raffine.shape)
                               + constr.raffine.const)
                    this_const = constr.affine.const
                    this_const = this_const.flatten()
                    roaffine = RoAffine(raffine, aff_linear@drule + this_const,
                                        self.sup_model)
                    ew_constr = RoConstr(roaffine, constr.sense)
                else:
                    raise TypeError('Unknown type.')

            elif isinstance(constr, DecLinConstr):
                linear = constr.linear
                const = constr.const
                roaffine = linear @ drule - const.reshape(const.size)
                if isinstance(roaffine, RoAffine):
                    is_equal = (all(constr.sense) if
                                isinstance(constr.sense, Iterable) else
                                constr.sense == 1)
                    if is_equal == 1:

                        left = DecLinConstr(constr.model,
                                            constr.linear, constr.const,
                                            np.zeros(constr.linear.shape[0]),
                                            constr.event_adapt, constr.ctype)
                        right = DecLinConstr(constr.model,
                                             -constr.linear, -constr.const,
                                             np.zeros(constr.linear.shape[0]),
                                             constr.event_adapt, constr.ctype)
                        return self.ro_to_roc(left) + self.ro_to_roc(right)

                    if (roaffine.raffine.linear.nnz == 0 and
                        not roaffine.raffine.const.any()):
                        ew_constr = LinConstr(roaffine.dec_model,
                                              roaffine.affine.linear,
                                              - roaffine.affine.const,
                                              constr.sense)
                    else:
                        ew_constr = RoConstr(roaffine, constr.sense)
                elif isinstance(roaffine, Affine):
                    ew_constr = LinConstr(roaffine.model, roaffine.linear,
                                          -roaffine.const, constr.sense)
                else:
                    raise TypeError('Unknown type.')
            elif isinstance(constr, DecCvxConstr):
                linear_in = constr.affine_in.linear
                const_in = constr.affine_in.const
                aff_in = linear_in@drule + const_in.reshape(const_in.size)
                if isinstance(aff_in, RoAffine):
                    if (aff_in.raffine.linear.nnz > 0 or
                            np.any(aff_in.raffine.const)):
                        raise SyntaxError('Incorrect convex expressions.')
                    aff_in = aff_in.affine
                linear_out = constr.affine_out.linear
                const_out = constr.affine_out.const
                aff_out = linear_out@drule + const_out.reshape(const_out.size)
                if isinstance(aff_out, RoAffine):
                    if (aff_out.raffine.linear.nnz > 0 or
                            np.any(aff_out.raffine.const)):
                        raise SyntaxError('Incorrect convex expressions.')
                    aff_out = aff_out.affine
                ew_constr = CvxConstr(aff_in.model, aff_in, aff_out,
                                      constr.xtype)
            else:
                raise TypeError('Unknown constraint type.')

            if isinstance(ew_constr, RoConstr):
                if (ew_constr.raffine.linear.nnz > 0 or
                        np.any(ew_constr.raffine.const)):
                    if constr.ambset is None:
                        if self.obj_ambiguity is None:
                            raise SyntaxError('The Ambiguity set is '
                                              'undefined.')
                        else:
                            ambset = self.obj_ambiguity
                            support = ambset.sup_constr[s]
                    else:
                        ambset = constr.ambset
                        if isinstance(ambset, Ambiguity):
                            support = constr.ambset.sup_constr[s]
                        elif isinstance(ambset, Iterable):
                            support = ambset
                    ew_constr = ew_constr.forall(support)
                else:
                    ew_constr = LinConstr(ew_constr.affine.model,
                                          ew_constr.affine.linear,
                                          ew_constr.affine.const,
                                          ew_constr.sense)

            ro_constr.append(ew_constr)

        return ro_constr

    def dro_to_roc(self, constr):

        drule_list = self.rule_var()
        num_var = self.vt_model.vars[-1].last
        num_scen = self.num_scen
        num_rand = self.sup_model.vars[-1].last

        # Ambiguity set of the constr
        ambset = constr.ambset if constr.ambset else self.obj_ambiguity
        if not ambset:
            raise ValueError('The ambiguity set is undefined.')
        mixed_support = ambset.mix_support(primal=False)
        p = ambset.mix_model.vars[0][:num_scen]
        # p = ambset.mix_model.vars[0]
        var_exp_list = ambset.mix_model.vars[1:]
        num_event = len(ambset.exp_constr)

        # Constraint components
        if isinstance(constr, DecLinConstr):
            linear = constr.linear
            const = constr.const
            raffine = None
        else:
            linear = constr.affine.linear
            const = constr.affine.const
            raffine = constr.raffine

        # Standardize constraints
        ro_constr = []
        for i in range(linear.shape[0]):
            alpha = self.ro_model.dvar(num_scen)
            if num_event:
                beta = self.ro_model.dvar((num_rand, num_event))
            else:
                beta = 0

            # left = alpha @ p[:num_scen]
            left = alpha @ p
            for j in range(num_event):
                left += var_exp_list[j][:num_rand] @ beta[:, j]
            ro_constr.extend((left <= 0).le_to_rc(mixed_support))

            z = Vars(self.sup_model, 0, (num_rand,), 'C', None)
            for s in range(num_scen):
                drule = drule_list[s]
                left = linear[i, :num_var] @ drule + const[i]
                if raffine:
                    if isinstance(drule, RoAffine):
                        extra = left.raffine
                        temp = drule.affine
                        left = left.affine.reshape(left.shape)
                    elif isinstance(drule, Affine):
                        extra = 0
                        temp = drule
                    else:
                        raise TypeError('Incorrect data type.')
                    row_ind = i*num_rand + np.arange(num_rand, dtype=int)
                    new_raffine = raffine.linear[row_ind] @ temp
                    new_raffine = new_raffine.reshape((1, new_raffine.size))
                    new_raffine += raffine.const[i, :num_rand] + extra
                    left = RoAffine(new_raffine, left, constr.rand_model)  # ##

                event_indices = [k for k in range(num_event)
                                 if s in ambset.exp_constr_indices[k]]
                if len(event_indices) > 0:
                    right = alpha[s] + (z @ beta[:, event_indices]).sum()
                else:
                    right = alpha[s]
                inequality = (left <= right)
                if isinstance(inequality, RoConstr):
                    ro_constr.append(inequality.forall(ambset.sup_constr[s]))
                elif isinstance(inequality, LinConstr):
                    ro_constr.append(inequality)
                else:
                    raise TypeError('Incorrect data type.')

        return ro_constr

    def solve(self, solver=None, display=True, export=False, params={}):
        """
        Solve the model with the selected solver interface.

        Parameters
        ----------
            solver : {None, lpg_solver, grb_solver, msk_solver}
                Solver interface used for model solution. Use default solver
                if solver=None.
            display : bool
                Display option of the solver interface.
            export : bool
                Export option of the solver interface. A standard model file
                is generated if the option is True.
            params : dict
                A dictionary that specifies parameters of the selected solver.
                So far the argument only applies to Gurobi and MOSEK.
        """

        if solver is None:
            solution = def_sol(self.do_math(), display, export, params)
        else:
            solution = solver.solve(self.do_math(), display, export, params)

        if isinstance(solution, Solution):
            self.ro_model.solution = solution
        else:
            if not solution:
                warnings.warn('No feasible solutions can be found.')
            else:
                x = solution.x
                self.ro_model.solution = Solution(x[0], x, solution.status)

        self.ro_model.rc_model.solution = solution
        self.solution = solution

    def get(self):

        if self.solution is None:
            raise SyntaxError('The model is unsolved or no feasible solution.')
        return self.sign * self.solution.objval


class Ambiguity:

    def __init__(self, model):

        self.model = model

        self.sup_constr = [None] * model.num_scen
        self.exp_constr = []
        self.exp_constr_indices = []
        self.mix_model = None

        p = self.model.p
        self.pro_constr = [p >= 0, sum(p) == 1]
        self.s = Scen(self, self.model.series_scen, self.model.p)

        self.update = True

    def __str__(self):

        return self.s.__str__()

    def __repr__(self):

        return self.s.__repr__()

    def showevents(self):

        num_scen = self.model.num_scen
        table = pd.DataFrame(['undefined']*num_scen,
                             columns=['support'],
                             index=self.s.series.index)
        sup_constr = self.sup_constr
        defined = pd.notnull(pd.Series(sup_constr,
                                       index=self.s.series.index))
        table.loc[defined, 'support'] = 'defined'

        # exp_cosntr = self.model.exp_constr
        exp_constr_indices = self.exp_constr_indices
        count = 0
        for indices in exp_constr_indices:
            column = 'expectation {0}'.format(count)
            count += 1
            table[column] = False
            table.iloc[indices, count] = True

        return table

    def __getitem__(self, indices):

        return self.s[indices]

    @property
    def loc(self):

        return self.s.loc

    @property
    def iloc(self):

        return self.s.iloc

    def suppset(self, *args):

        self.update = True
        return self.s.suppset(*args)

    def exptset(self, *args):

        self.update = True
        return self.s.exptset(*args)

    def probset(self, *args):

        self.update = True
        for arg in args:
            if arg.model is not self.model.pro_model:
                raise ValueError('Constraints are not defined for the ' +
                                 'probability set.')

        pr = self.model.p
        self.pro_constr = [pr >= 0, pr.sum() == 1] + list(args)

    def mix_support(self, primal=True):

        if not self.update and self.mix_model is not None:
            return self.mix_model.do_math(primal)

        self.model.pro_model.reset()
        self.model.pro_model.st(self.pro_constr)
        pro_support = self.model.pro_model.do_math()
        self.mix_model = SOCModel(nobj=True, mtype='M')

        # Constraints for probabilities
        p = self.mix_model.dvar(pro_support.linear.shape[1])
        constr = LinConstr(self.mix_model,
                           pro_support.linear, pro_support.const,
                           pro_support.sense)
        self.mix_model.st(constr)
        for q in pro_support.qmat:
            cconstr = ConeConstr(self.mix_model, p, q[1:], p, q[0])
            self.mix_model.st(cconstr)

        # Constraints for expectations
        for econstr, indices in zip(self.exp_constr, self.exp_constr_indices):
            self.model.exp_model.reset()
            self.model.exp_model.st(econstr)
            exp_support = self.model.exp_model.do_math()
            exp_var = self.mix_model.dvar(exp_support.linear.shape[1])
            affine = (exp_support.linear @ exp_var
                      - p[indices].sum() * exp_support.const)
            constr = LinConstr(affine.model, affine.linear, affine.const,
                               exp_support.sense)
            self.mix_model.st(constr)
            for q in exp_support.qmat:
                cconstr = ConeConstr(self.mix_model, exp_var, q[1:],
                                     exp_var, q[0])
                self.mix_model.st(cconstr)

        return self.mix_model.do_math(primal)
